/* $XConsortium: miscanfillZ.h,v 1.2 93/08/09 19:45:47 takamori Exp $ */
#ifndef ZSCANFILLINCLUDED
#define ZSCANFILLINCLUDED
/*
 *     scanfillZ.h
 *
 *     Written by Brian Kelleher; Jan 1985
 *
 *     This file contains a few macros to help track
 *     the edge of a filled object.  The object is assumed
 *     to be filled in scanline order, and thus the
 *     algorithm used is an extension of Bresenham's line
 *     drawing algorithm which assumes that y is always the
 *     major axis.
 *     Since these pieces of code are the same for any filled shape,
 *     it is more convenient to gather the library in one
 *     place, but since these pieces of code are also in
 *     the inner loops of output primitives, procedure call
 *     overhead is out of the question.
 *     See the author for a derivation if needed.
 */


/*
 *  In scan converting polygons, we want to choose those pixels
 *  which are inside the polygon.  Thus, we add .5 to the starting
 *  x coordinate for both left and right edges.  Now we choose the
 *  first pixel which is inside the pgon for the left edge and the
 *  first pixel which is outside the pgon for the right edge.
 *  Draw the left pixel, but not the right.
 *
 *  How to add .5 to the starting x coordinate:
 *      If the edge is moving to the right, then subtract dy from the
 *  error term from the general form of the algorithm.
 *      If the edge is moving to the left, then add dy to the error term.
 *
 *  The reason for the difference between edges moving to the left
 *  and edges moving to the right is simple:  If an edge is moving
 *  to the right, then we want the algorithm to flip immediately.
 *  If it is moving to the left, then we don't want it to flip until
 *  we traverse an entire pixel.
 */
#define ZBRESINITPGON(dy, x1, x2, z1, z2, xStart, zStart, dd, ddz, \
		     m, m1, mz, mz1, incr1, incr2, incrz1, incrz2) { \
    int dx, dz;      /* local storage */ \
\
    /* \
     *  if the edge is horizontal, then it is ignored \
     *  and assumed not to be processed.  Otherwise, do this stuff. \
     */ \
    if ((dy) != 0) { \
        xStart = (x1); \
        zStart = (z1); \
        dx = (x2) - xStart; \
        dz = (z2) - zStart; \
        if (dx < 0) { \
            m = dx / (dy); \
            m1 = m - 1; \
            incr1 = -2 * dx + 2 * (dy) * m1; \
            incr2 = -2 * dx + 2 * (dy) * m; \
            dd = 2 * m * (dy) - 2 * dx - 2 * (dy); \
        } else { \
            m = dx / (dy); \
            m1 = m + 1; \
            incr1 = 2 * dx - 2 * (dy) * m1; \
            incr2 = 2 * dx - 2 * (dy) * m; \
            dd = -2 * m * (dy) + 2 * dx; \
        } \
        if (dz < 0) { \
            mz = dz / (dy); \
            mz1 = mz - 1; \
            incrz1 = -2 * dz + 2 * (dy) * mz1; \
            incrz2 = -2 * dz + 2 * (dy) * mz; \
            ddz = 2 * mz * (dy) - 2 * dz - 2 * (dy); \
        } else { \
            mz = dz / (dy); \
            mz1 = mz + 1; \
            incrz1 = 2 * dz - 2 * (dy) * mz1; \
            incrz2 = 2 * dz - 2 * (dy) * mz; \
            ddz = -2 * mz * (dy) + 2 * dz; \
        } \
    } \
}

#define ZBRESINCRPGON(d, minval, m, m1, incr1, incr2) { \
    if (m1 > 0) { \
        if (d > 0) { \
            minval += m1; \
            d += incr1; \
        } \
        else { \
            minval += m; \
            d += incr2; \
        } \
    } else {\
        if (d >= 0) { \
            minval += m1; \
            d += incr1; \
        } \
        else { \
            minval += m; \
            d += incr2; \
        } \
    } \
}

/*
 *     This structure contains all of the information needed
 *     to run the bresenham algorithm.
 *     The variables may be hardcoded into the declarations
 *     instead of using this structure to make use of
 *     register declarations.
 */
typedef struct {
    int minor;         /* minor axis        */
    int minorz;        /* z axis        */
    int d;           /* decision variable */
    int dz;           /* decision variable */
    int m, m1;       /* slope and slope+1 */
    int mz, mz1;       /* zslope and zslope+1 */
    int incr1, incr2; /* error increments */
    int incrz1, incrz2; /* z-error increments */
} ZBRESINFO;

#define ZBRESINITPGONSTRUCT(dmaj, min1, min2, minz1, minz2, bres) \
        ZBRESINITPGON(dmaj, min1, min2, minz1, minz2, bres.minor, bres.minorz, bres.d, bres.dz, bres.m, bres.m1, bres.mz, bres.mz1, bres.incr1, bres.incr2, bres.incrz1, bres.incrz2)

#define ZBRESINCRPGONSTRUCT(bres) \
        ZBRESINCRPGON(bres.d, bres.minor, bres.m, bres.m1, bres.incr1, bres.incr2) \
        ZBRESINCRPGON(bres.dz, bres.minorz, bres.mz, bres.mz1, bres.incrz1, bres.incrz2)


#endif
